/*   Copyright (c) AT&T Corp.  All rights reserved.
   
This software may only be used by you under license from 
AT&T Corp. ("AT&T").  A copy of AT&T's Source Code Agreement 
is available at AT&T's Internet website having the URL 

http://www.research.att.com/sw/tools/graphviz/license/

If you received this software without first entering into a license 
with AT&T, you have an infringing copy of this software and cannot 
use it without violating AT&T's intellectual property rights. */

// overload >> DString for better-delimited tokens
// returns one token or one delimeter
inline istream &operator>>(istream &in,DString &s) {
	static const string delim = "=,[]; \t";
	string temp;
	in >> ws;
	while(!in.eof()) {
		char c = in.get();
		if(c<0)
			break; // why does eof() delay?
		unsigned i;
		for(i=0; i<delim.length(); ++i)
			if(delim[i]==c) {
				if(temp.empty())
					temp += c;
				else
					in.putback(c);
				break;
			}
		if(i<delim.length())
			break;
		temp += c;
	}
	s = temp.c_str();
	return in;
}
struct ParseExpected : DGException { 
  DString expect,got; 
  ParseExpected(DString expect,DString got) : expect(expect),got(got),DGException("the string parser didn't get what it expected") {} 
};
struct must {
	DString s;
	must(DString s) : s(s) {}
};
inline istream &operator>>(istream &in,const must &m) {
	DString s;
	in >> s;
	if(s!=m.s)
		throw ParseExpected(m.s,s);
	return in;
}
struct quote {
	const string &s;
	quote(const string &s) : s(s) {}
};
inline ostream &operator<<(ostream &out,quote &o) {
  out << '"';
  if(!o.s.empty()) 
    for(unsigned i = 0; i<o.s.length(); ++i)
      if(o.s[i]=='"') // " => \"
	out << "\\\""; 
      else if(o.s[i]=='\\') // \ => \\       .
	out << "\\\\";
      else
	out << o.s[i];
  out << '"';
  return out;
}
struct unquote {
	string &s;
	unquote(string &s) : s(s) {}
};
inline istream &operator>>(istream &in,unquote &o) {
	char c;
	in >> ws;
	in.get(c);
	if(c!='"') {
		char bf[2];
		bf[0] = c,bf[1] = 0;
		throw ParseExpected("\"",bf);
	}
	o.s.erase();
	string temp;
	while(1) {
		in.get(c);
		if(c=='\\')
			in.get(c);
		else if(c=='"')
			break;
		o.s += c;
	}
	return in;
}
